#include <math.h>
#include <clutter/clutter.h>
#include <clutter-mozembed.h>
#include "odo-distort-funcs.h"
#include "odo-texture.h"

typedef struct
{
  ClutterActor    *odo;
  OdoDistortData   data;
  ClutterTimeline *timeline;
  ClutterAlpha    *alpha;
  gboolean         fade_out;
} DistortData;

static void
completed_cb (ClutterTimeline *timeline,
              DistortData     *data)
{
  clutter_container_remove_actor (
    CLUTTER_CONTAINER (clutter_actor_get_parent (data->odo)), data->odo);
  g_object_unref (data->alpha);
  g_object_unref (data->timeline);
}

static void
new_frame_cb (ClutterTimeline *timeline,
              gint             msecs,
              DistortData     *data)
{
  /* Set the turn value to the alpha value and redraw */
  data->data.turn = clutter_alpha_get_alpha (data->alpha);
  data->data.amplitude = data->data.turn;

  odo_texture_invalidate (ODO_TEXTURE (data->odo));
  clutter_actor_set_opacity (data->odo,
                             (guint8)((1.f-data->data.turn) * 255.f));
}

static void
notify_location_cb (ClutterMozEmbed *page,
                    GParamSpec      *pspec,
                    ClutterText     *entry)
{
  static gboolean first_run = TRUE;

  clutter_text_set_text (entry, clutter_mozembed_get_location (page));

  if (!first_run)
    {
      CoglHandle old_cogl_tex/*, new_cogl_tex*/;
      ClutterContainer *container;
      ClutterTexture *texture;
      gint width, height;
      DistortData *distort_data;
      /*guchar *data;
      gint size;*/

      old_cogl_tex = clutter_texture_get_cogl_texture (CLUTTER_TEXTURE (page));
      clutter_texture_get_base_size (CLUTTER_TEXTURE (page), &width, &height);

      /* Bit of a hack, we know that the entire page will change when
       * we go to a new URL, so the texture handle will change - so we just
       * ref the old one instead of copying.
       */
      /*size = cogl_texture_get_data (old_cogl_tex,
                                    COGL_PIXEL_FORMAT_BGRA_8888_PRE,
                                    0,
                                    NULL);

      if (!size)
        return;

      data = g_malloc (size);
      new_cogl_tex =
        cogl_texture_new_from_data (width, height,
                                    COGL_TEXTURE_NO_SLICING,
                                    COGL_PIXEL_FORMAT_BGRA_8888_PRE,
                                    COGL_PIXEL_FORMAT_BGRA_8888_PRE,
                                    width * 4,
                                    data);
      g_free (data);*/

      container = CLUTTER_CONTAINER (clutter_actor_get_parent (
                    CLUTTER_ACTOR (page)));

      /* Set the COGL texture */
      texture = CLUTTER_TEXTURE (clutter_texture_new ());
      clutter_texture_set_cogl_texture (texture, old_cogl_tex);
      //cogl_handle_unref (new_cogl_tex);

      distort_data = g_new0 (DistortData, 1);
      distort_data->odo = odo_texture_new_with_textures (texture, NULL);
      odo_texture_set_callback (ODO_TEXTURE (distort_data->odo),
                                page_turn_func,
                                &distort_data->data);
      distort_data->timeline = clutter_timeline_new (800);
      distort_data->alpha = clutter_alpha_new_full (distort_data->timeline,
                                            CLUTTER_EASE_IN_OUT_QUAD);
      distort_data->data.radius = width/14.f;
      distort_data->data.angle = G_PI/6.f;
      distort_data->data.amplitude = 1.f;

      g_signal_connect (distort_data->timeline, "new-frame",
                        G_CALLBACK (new_frame_cb), distort_data);
      g_signal_connect (distort_data->timeline, "completed",
                        G_CALLBACK (completed_cb), distort_data);

      clutter_actor_set_position (distort_data->odo,
                                  clutter_actor_get_x (CLUTTER_ACTOR (page)),
                                  clutter_actor_get_y (CLUTTER_ACTOR (page)));
      clutter_container_add_actor (container, distort_data->odo);

      clutter_timeline_start (distort_data->timeline);
    }
  else
    first_run = FALSE;
}

static void
entry_activate_cb (ClutterText     *text,
                   ClutterMozEmbed *page)
{
  clutter_mozembed_open (page, clutter_text_get_text (text));
  clutter_actor_grab_key_focus (CLUTTER_ACTOR (page));
}

static void
net_start_cb (ClutterMozEmbed *page, ClutterActor *load_rect)
{
  clutter_actor_set_width (load_rect, 0);
  clutter_actor_animate (load_rect, CLUTTER_EASE_IN_QUAD, 250,
                         "opacity", 0xff,
                         NULL);
}

static void
net_stop_cb (ClutterMozEmbed *page, ClutterActor *load_rect)
{
  clutter_actor_animate (load_rect, CLUTTER_EASE_IN_QUAD, 250,
                         "opacity", 0x00,
                         NULL);
}

static gboolean
grab_key_focus (ClutterActor *actor)
{
  clutter_actor_grab_key_focus (actor);
  return FALSE;
}

static void
progress_cb (ClutterMozEmbed *page,
             gdouble          progress,
             ClutterActor    *load_rect)
{
  clutter_actor_set_width (load_rect, CLUTTER_STAGE_WIDTH () * progress);
}

int
main (int argc, char **argv)
{
  gfloat width, height, text_height, page_width, page_height;

  const ClutterColor background = { 0xcc, 0xcc, 0xcc, 0xff };
  const ClutterColor select = { 0x6f, 0xaa, 0xff, 0xff };
  const ClutterColor base = { 0xff, 0xff, 0xff, 0xff };
  const ClutterColor text = { 0x00, 0x00, 0x00, 0xff };

  ClutterActor *stage, *entry, *load_rect, *bg_rect, *group, *page, *reflection;

  /* Need to call gtk_init if cluttermozembed built with plugins */
  gtk_init (&argc, &argv);

  clutter_init (&argc, &argv);

  /* Create all the actors */
  stage = clutter_stage_get_default ();
  clutter_stage_set_color (CLUTTER_STAGE (stage), &background);

  entry = clutter_text_new_full ("Sans 16", "Loading...", &text);
  clutter_text_set_editable (CLUTTER_TEXT (entry), TRUE);
  clutter_text_set_single_line_mode (CLUTTER_TEXT (entry), TRUE);

  bg_rect = clutter_rectangle_new_with_color (&base);

  load_rect = clutter_rectangle_new_with_color (&select);

  group = clutter_group_new ();

  page = clutter_mozembed_new ();

  reflection = clutter_clone_new (group);

  /* Connect signals */
  g_signal_connect (page, "notify::location",
                    G_CALLBACK (notify_location_cb), entry);
  g_signal_connect (page, "net-start",
                    G_CALLBACK (net_start_cb), load_rect);
  g_signal_connect (page, "net-stop",
                    G_CALLBACK (net_stop_cb), load_rect);
  g_signal_connect (page, "progress",
                    G_CALLBACK (progress_cb), load_rect);
  g_signal_connect (page, "button-press-event",
                    G_CALLBACK (grab_key_focus), NULL);
  g_signal_connect (entry, "activate",
                    G_CALLBACK (entry_activate_cb), page);
  g_signal_connect (entry, "button-press-event",
                    G_CALLBACK (grab_key_focus), NULL);
  g_signal_connect_swapped (bg_rect, "button-press-event",
                            G_CALLBACK (grab_key_focus), page);
  g_signal_connect_swapped (load_rect, "button-press-event",
                            G_CALLBACK (grab_key_focus), page);

  /* Put it all together */
  clutter_stage_set_fullscreen (CLUTTER_STAGE (stage), TRUE);
  clutter_actor_get_size (stage, &width, &height);

  text_height = clutter_actor_get_height (entry);
  clutter_actor_set_position (entry, 0, height - text_height);
  clutter_actor_set_reactive (entry, TRUE);

  clutter_actor_set_size (bg_rect, width, text_height);
  clutter_actor_set_position (bg_rect, 0, height - text_height);
  clutter_actor_set_reactive (bg_rect, TRUE);

  clutter_actor_set_opacity (load_rect, 0x00);
  clutter_actor_set_size (load_rect, 0, text_height);
  clutter_actor_set_position (load_rect, 0, height - text_height);
  clutter_actor_set_reactive (load_rect, TRUE);

  page_width = width - 200;
  page_height = height - text_height - 200;
  clutter_actor_set_size (page, page_width, page_height);
  clutter_actor_set_position (page,
                              (width - page_width)/2,
                              (height - text_height - page_height)/2);

  clutter_actor_set_position (reflection,
                              0,
                              page_height + clutter_actor_get_y (page));
  clutter_actor_set_rotation (reflection, CLUTTER_X_AXIS, 180.0,
                              page_width / 2,
                              page_height / 2,
                              0);
  clutter_actor_set_opacity (reflection, 0x80);

  /* Add everything to the group/stage */
  clutter_container_add_actor (CLUTTER_CONTAINER (group), page);
  clutter_container_add (CLUTTER_CONTAINER (stage),
                         reflection,
                         group,
                         bg_rect,
                         load_rect,
                         entry,
                         NULL);

  /* Load a page */
  if (argc > 1)
    clutter_mozembed_open (CLUTTER_MOZEMBED (page), argv[1]);
  else
    clutter_mozembed_open (CLUTTER_MOZEMBED (page), "http://www.intel.com/");

  /* Start */
  clutter_actor_grab_key_focus (entry);
  clutter_actor_show (stage);
  clutter_main ();

  return 0;
}

